/*
 * DolmenApp.h
 *
 * Copyright (C) 2010-2013 Julien Eychenne
 *
 * This file is part of Dolmen.
 *
 * Dolmen is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License (GPL) as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * Dolmen is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Dolmen. If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:  Julien Eychenne
 * Purpose: Main app
 * Created: September 2010
 */

#ifndef DOLMENAPP_H
#define DOLMENAPP_H

#include <QMainWindow>
#include <QVBoxLayout>
#include <QSplitter>
#include <QSizePolicy>
#include <QMessageBox>
#include <QSpacerItem>
#include <QPushButton>
#include <QSet>
#include <QString>
#include <QToolBar>
#include <QTime>
#include <QProgressDialog>
#include <QProcess>

#include "lib/Project.h"
#include "lib/Plugin.h"
#include "lib/SqlDatabase.h"
#include "lib/ZipArchive.h"
#include "utils/routines.h"
#include "utils/Global.h"
#include "search/QueryParser.h"
#include "api/DMScriptApi.h"
#include "gui/FileSidebar.h"
#include "gui/MetadataSidebar.h"
#include "gui/FileBrowser.h"
#include "gui/BookmarkBrowser.h"
#include "gui/Viewer.h"
#include "gui/BottomPane.h"
#include "gui/OutputTab.h"
#include "gui/MainMenu.h"
#include "gui/actions.h"
#include "gui/QueryView.h"
#include "gui/SearchWindow.h"
#include "gui/splitters.h"
#include "gui/PreferenceEditor.h"
#include "gui/GaugeDialog.h"
#include "gui/GaugeView.h"
#include "gui/AnnotationView.h"





class DolmenApp : public QMainWindow
{
    Q_OBJECT

public:
	DolmenApp(QString version, QWidget *parent = 0, const QString project = QString());
	~DolmenApp();

public slots:
	void onMainSearchClicked();
	void warnUser(QString &msg);
	void displayQuery(Query *);
	void closeEvent(QCloseEvent *);
	void askPath(DFile *);

	void displayImportProgressBar(int);
	void updateImportProgressBar(int);
	void loadQuery(const QString &);
	void forwardQuery(int);
	// Dolmen's standard output
	void output(const QString &text);
	void newOutput(const QString &text); // clear before writing
	// standard error
	void error(const QString &text);
	void about();
	void license();
	void acknowledgements();


signals:
	void app_saveFileWithPath(DFile *, const QString &);
	void app_openProject();

private slots:
	void onFileBrowserDoubleClicked();
	void closeCurrentView();
	void openRecentProject();
	void editPreferences();
	void editProject();
	void updatePreferences();
	void enableProjectActions(); // enable actions such as "Save project...", "Open query..."
	void openLastProject();
	void setGlobalProjectPath(QString);
	void clearRecentProjects();
	void openDocumentation();
	void cleanUpProcess();
	void displaySupportedAudioFormats();
	void toggleFileBrowser();
	void toggleBottomTabs();
    void toggleMetadata();
	void compareSpeakers();
	void openGaugeView(const QString &task, int tier_no);
	void openAnnotationMatch(Annotation *a, int tier, double tstart, double tend);
    void installUserPlugin();
    void importMetadata(const QString &path);

private:
    QWidget             *root_widget, *wgt_browser_bar;
    QVBoxLayout         *vlayout_browser_bar;
    Splitter            *hsplitter, *vsplitter, *view_splitter;
    QPushButton         *btn_main_search;
    QToolBar            *toolbar;
    QMenu               *projects_submenu;
    QProgressDialog     *import_progress;
    QString              version;
    BrowserSideBar      *sidebar;
    FileBrowser         *file_browser;
    BookmarkBrowser     *bookmark_browser;
    Viewer              *viewer;
    BottomPane          *tabs_bottom;
    PreferenceEditor    *prefs_editor;
    MetadataSideBar     *metadata_widget;
    MainMenu            *menu;
    QMenu               *file_menu, *edit_menu, *window_menu, *tools_menu, *help_menu;
    Project             *current_project;
    QueryParser         *query_parser;

    ActionOpenProject   *action_open_project;
    ActionAddFile       *action_add_file;
    ActionOpenQuery     *action_open_query;
    ActionAddFolder     *action_add_folder;
    ActionSaveProject   *action_save_project;
    ActionSaveProjectAs *action_save_project_as;
    ActionImportMetaCsv *action_import_meta;
    QAction             *action_close_view, *action_show_browser, *action_show_bottom_tabs,
                        *action_show_metadata, *action_compare_speakers, *action_compare_annotators;


    QProcess            *process; // process to display doc
    QList<Query*>        queries; // keep track of all previous queries, in order
    MainDatabase        *m_db;
    PraatInstance       *praat;
    DMScriptApi         *script_api;

	// initialize GUI
	void setProject();
	void setupUi();
	void setMainMenu();
	void setFileMenu();
	void setEditMenu();
	void setWindowMenu();
	void setToolsMenu();
	void setHelpMenu();
    void setBrowserBar();
    void setMetadataBar();
	void setSplitters();
    void setBrowserBarLayout();
	void setViewer();
	void setBottomTabs();
	void setConnections();
	void checkPraat();
	void resetQueryParser();
	void readSettings();
	void writeSettings();
    void setQueryViewConnections(QueryView *view);
	void initializeSound();
    void initializePlugins();
    void initializeGrammars();
    void testPlugins(QDir root, QStringList &plugins);
    void setMainPlugin(Plugin *plugin);
	void closeApplication();
    void setMainDatabase();
};

#endif // DOLMENAPP_H
